#include "music.h"
#include "m_opl.h"

#ifdef linux
static __inline__ BYTE inportb(const WORD port)
{
   BYTE _v;
   __asm__ volatile ("inb %1,%0" : "=a" (_v):"d" ((WORD) port));
   return _v;
}
static __inline__ void outportb(const WORD port, const BYTE value)
{
   __asm__ volatile ("outb %0,%1" ::"a" ((BYTE)value),"d" ((WORD)port));
}
#ifndef MAKEDEPEND
 #include <unistd.h>
 #include <sys/io.h>
#endif
#else
#ifndef MAKEDEPEND
 #include <conio.h>
#endif
#endif

#define TEST_REGISTER                0x01
#define   ENABLE_WAVE_SELECT           0x20

#define TIMER1_REGISTER              0x02
#define TIMER2_REGISTER              0x03
#define TIMER_CONTROL_REGISTER       0x04    /* Left side */
#define   IRQ_RESET                    0x80
#define   TIMER1_MASK                  0x40
#define   TIMER2_MASK                  0x20
#define   TIMER1_START                 0x01
#define   TIMER2_START                 0x02

#define CONNECTION_SELECT_REGISTER   0x04    /* Right side */
#define   RIGHT_4OP_0                  0x01
#define   RIGHT_4OP_1                  0x02
#define   RIGHT_4OP_2                  0x04
#define   LEFT_4OP_0                   0x08
#define   LEFT_4OP_1                   0x10
#define   LEFT_4OP_2                   0x20

#define OPL3_MODE_REGISTER           0x05    /* Right side */
#define   OPL3_ENABLE                  0x01
#define   OPL4_ENABLE                  0x02

#define KBD_SPLIT_REGISTER           0x08    /* Left side */
#define   COMPOSITE_SINE_WAVE_MODE     0x80  /* Don't use with OPL-3? */
#define   KEYBOARD_SPLIT               0x40

#define PERCOSSION_REGISTER    0xBD    /* Left side only */
#define   TREMOLO_DEPTH          0x80
#define   VIBRATO_DEPTH          0x40
#define   PERCOSSION_ENABLE      0x20
#define   BASSDRUM_ON            0x10
#define   SNAREDRUM_ON           0x08
#define   TOMTOM_ON              0x04
#define   CYMBAL_ON              0x02
#define   HIHAT_ON               0x01

#define AM_VIB                 0x20
#define   TREMOLO_ON             0x80
#define   VIBRATO_ON             0x40
#define   SUSTAIN_ON             0x20
#define   KSR                    0x10    /* Key scaling rate */
#define   MULTIPLE_MASK          0x0F    /* Frequency multiplier */

#define KSL_LEVEL              0x40
#define   KSL_MASK               0xC0    /* Envelope scaling bits */
#define   TOTAL_LEVEL_MASK       0x3F    /* Strength (volume) of OP */

#define ATTACK_DECAY           0x60
#define   ATTACK_MASK            0xF0
#define   DECAY_MASK             0x0F

#define SUSTAIN_RELEASE        0x80
#define   SUSTAIN_MASK           0xF0
#define   RELEASE_MASK           0x0F

#define WAVE_SELECT            0xE0

#define FNUM_LOW               0xA0
#define KEYON_BLOCK            0xB0
#define   KEYON_BIT              0x20
#define   BLOCKNUM_MASK          0x1C
#define   FNUM_HIGH_MASK         0x03
#define FEEDBACK_CONNECTION    0xC0
#define   FEEDBACK_MASK          0x0E  /* Valid just for 1st OP of a voice */
#define   CONNECTION_BIT         0x01
#define   STEREO_BITS            0x30    /* OPL-3 only */
#define     VOICE_TO_LEFT         0x10
#define     VOICE_TO_RIGHT        0x20

static const WORD OPLBase = 0x388;
static const BYTE Portit[9] = {0,1,2, 8,9,10, 16,17,18};
static const WORD Periods[12] = {1712,1616,1524,1440,
								 1356,1280,1208,1140,
								 1076,1016, 960, 907};

static long Period[9];
static WORD C4Spd[9];
static BYTE *InsD[9];
static SBYTE Pan[9];
static BYTE cnote[9];

#ifdef linux
static int did_ioperm=0;
#endif

static void OPL_Byte(BYTE Index, BYTE Data)
{
	register int a;
	outportb(OPLBase, Index);  for(a=0; a<6;  a++)inportb(OPLBase);
	outportb(OPLBase+1, Data); for(a=0; a<35; a++)inportb(OPLBase);
}
static void OPL_Reset(void)
{
	int a;
	for(a=0; a<244; OPL_Byte(a++, 0));
	for(a=0; a<9; a++)
	{
		C4Spd[a] = 8363;
		Pan[a] = 0;
	}
}
static void OPL_Close(void)
{
	#ifdef linux
	if(!did_ioperm)return;
	#endif
	if(OPLDriver.active)OPL_Reset();
	#ifdef linux
	ioperm(OPLBase, 4, 0);
	did_ioperm = 0;
	#endif
	OPLDriver.active = 0;
}
static int OPL_Detect(void)
{
	int st1, st2;
	
	#ifdef linux
	if(!did_ioperm)
	{
		if(ioperm(OPLBase, 4, 1) < 0)return -1;
		did_ioperm=1;
	}
	#endif
	
	OPL_Reset();
	
	OPL_Byte(TIMER_CONTROL_REGISTER, TIMER1_MASK|TIMER2_MASK);
	OPL_Byte(TIMER_CONTROL_REGISTER, IRQ_RESET);
	st1 = inportb(OPLBase) & 0xE0;
	OPL_Byte(TIMER1_REGISTER, 0xFF);
	OPL_Byte(TIMER_CONTROL_REGISTER, TIMER2_MASK|TIMER1_START);
	#ifdef __BORLANDC__
	 delay(10);
	#else
	 usleep(80);
	#endif
	st2 = inportb(OPLBase) & 0xE0;
	OPL_Byte(TIMER_CONTROL_REGISTER, TIMER1_MASK|TIMER2_MASK);
	OPL_Byte(TIMER_CONTROL_REGISTER, IRQ_RESET);
	
	OPLDriver.active = (st1==0 && st2==0xC0);
	
	if(!OPLDriver.active)OPL_Close();
	return OPLDriver.active - 1;
}
static void OPL_Patch(int ch, int noteadd, int midipatch,
					  int c4spd, BYTE *D)
{
	int ope;
	
	if(ch >= 9 || ch < 0 || !D)return;
	midipatch = midipatch;
	noteadd   = noteadd;
	
	C4Spd[ch] = c4spd;
	InsD[ch] = D;
	
	ope = Portit[ch];
	
	OPL_Byte(AM_VIB+           ope, D[0]);
	OPL_Byte(ATTACK_DECAY+     ope, D[4]);
	OPL_Byte(SUSTAIN_RELEASE+  ope, D[6]);
	OPL_Byte(WAVE_SELECT+      ope, D[8]&3); /* 6 high bits used elsewhere */

	OPL_Byte(AM_VIB+         3+ope, D[1]);
	OPL_Byte(ATTACK_DECAY+   3+ope, D[5]);
	OPL_Byte(SUSTAIN_RELEASE+3+ope, D[7]);
	OPL_Byte(WAVE_SELECT+    3+ope, D[9]&3); /* 6 high bits used elsewhere */

	/* Panning... */
    OPL_Byte(FEEDBACK_CONNECTION + ch,
        (D[10] & ~STEREO_BITS)
            | (Pan[ch]<-32 ? VOICE_TO_LEFT
                : Pan[ch]>32 ? VOICE_TO_RIGHT
                : (VOICE_TO_LEFT | VOICE_TO_RIGHT)
            ) );
}

static void OPL_Pan(int ch, int value)
{
	if(ch >= 0 && ch < 9)
		Pan[ch] = value;
}

static void OPL_SendHerz(int ch, long Herz)
{
	unsigned Oct;
	
	for(Oct=0; Herz>0x1FF; Oct++)Herz >>= 1;
	
    OPL_Byte(FNUM_LOW+ch, (BYTE)Herz);
    OPL_Byte(KEYON_BLOCK+ch, KEYON_BIT
                      | ((Herz>>8) & FNUM_HIGH_MASK)
                      | ((Oct<<2) & BLOCKNUM_MASK)
            );
}

static void OPL_Touch(int ch, int volume)
{
	int ope;
	if(ch < 0 || ch >= 9)return;
	
	if(!InsD[ch])return;
	
	ope = Portit[ch];
	
	OPL_Byte(KSL_LEVEL+  ope, (InsD[ch][2]&KSL_MASK) | (63 + (InsD[ch][2]&63) * volume / 63 - volume));
	OPL_Byte(KSL_LEVEL+3+ope, (InsD[ch][3]&KSL_MASK) | (63 + (InsD[ch][3]&63) * volume / 63 - volume));	
}

static void OPL_NoteOn(int ch, int note, int volume)
{
	if(ch < 0 || ch >= 9)return;
	
	cnote[ch] = note;
	
	Period[ch] = ((8363L*64 * Periods[note%12]) >> (note/12)) / C4Spd[ch];
	
	OPL_SendHerz(ch, 14317056 / Period[ch]);
	OPL_Touch(ch, volume);
}

static void OPL_NoteOff(int ch)
{
	if(ch >= 0 && ch < 9)
	{
		int ope = Portit[ch];
		/* KEYON_BLOCK+ch seems to not work alone?? */
		OPL_Byte(KEYON_BLOCK+ch, 0);
		OPL_Byte(KSL_LEVEL+  ope, 0xFF);
		OPL_Byte(KSL_LEVEL+3+ope, 0xFF);
		
		Pan[ch] = 0;
	}
}

static void OPL_MidiCtrl(int ch, BYTE ctrl, BYTE value)
{
	ch=ch, ctrl=ctrl, value=value;
}

static void OPL_Bend(int ch, int value)
{
	int n = cnote[ch], n1,n2;
	long p1, p2;
	
	while(value < 0) { value += 64; n--; }
	while(value >= 64) { value -= 64; n++; }
	
	n1 = n;
	n2 = n+1;
	
	if(C4Spd[ch])
	{
		long p;
		
		p1 = ((8363L*64 * Periods[n1%12]) >> (n1/12)) / C4Spd[ch];
		p2 = ((8363L*64 * Periods[n2%12]) >> (n2/12)) / C4Spd[ch];
		
		p = (p1+(p2-p1)*value/64);
		
		if(p)
			OPL_SendHerz(ch, 14317056 / p);
	}
}

static void OPL_Delay(unsigned us)
{
	usleep(us);
}

musicdriver OPLDriver =
{
	"OPL3/FM synthesis, SB/AdLib -like cards etc...",
	
	"This driver can be used with any Ad Lib-compatible FM music\n"
	"synthesizer card (including Sound Blaster compatibles) that\n"
	"does not rely on software TSRs for Ad Lib FM compatibility.\n"
	"Users of more sophisticated General MIDI-based cards should\n"
	"select the Ad Lib driver only if problems occur with the card's\n"
	"General MIDI driver, as the Ad Lib OPL chipset does not offer\n"
    "the superior sound quality available with more modern devices.",
    
    0,	/* active       */
    9,	/* max channels */
    
    'o', 	/* preferred option character */
    
    OPL_Detect,
    OPL_Reset,
    OPL_Close,
    OPL_NoteOn,
    OPL_NoteOff,
    OPL_Touch,
    OPL_Patch,
    OPL_Pan,
    OPL_MidiCtrl,
    OPL_Bend,
    OPL_Delay,
    
    NULL
};
